//======================================================================
// 
// Regsys.c
//
// Copyright (C) 1996 Mark Russinovich and Bryce Cogswell
//
// Hooks the registry by replacing registry related calls in the system
// service table with pointers to our own hook routines. Very simple
// yet very effective.
//
//======================================================================
#include "ntddk.h"
#include "stdarg.h"
#include "stdio.h"
#include "ioctlcmd.h"
#include "regsys.h"

//----------------------------------------------------------------------
//                           DEFINES
//----------------------------------------------------------------------
// print macro that only turns on when debugging is on
#if DBG
#define DbgPrint(arg) DbgPrint arg
#else
#define DbgPrint(arg) 
#endif

// macro for easy hook/unhook
#define SYSCALL(_number)	ServiceTable->ServiceTable[ nttable[_number] ]

// convenient mutex macros
#define MUTEX_TYPE      KMUTEX
#define MUTEX_INIT(v)   KeInitializeMutex( &v, 0 )
#define MUTEX_P(v)      KeWaitForMutexObject( &v, Executive, KernelMode, FALSE, NULL )
#define MUTEX_V(v)      KeReleaseMutex( &v, FALSE )

// number of predefined rootkeys
#define NUMROOTKEYS     4

//----------------------------------------------------------------------
//                         FORWARD DEFINES
//---------------------------------------------------------------------- 
NTSTATUS RegmonDispatch( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );
VOID     RegmonUnload( IN PDRIVER_OBJECT DriverObject );

//----------------------------------------------------------------------
//                         GLOBALS
//---------------------------------------------------------------------- 
// our user-inteface device object
PDEVICE_OBJECT          GUIDevice;

// is a GUI talking to us?
BOOLEAN                 GUIActive = FALSE;

// is registry hooked?
BOOLEAN                 RegHooked = FALSE;

// global error string
CHAR                    errstring[256];

// pointer to system global service table
PSRVTABLE               ServiceTable;

// system call numbers for NT 3.51 and NT 4.0
ULONG	nt351[] = { 0x4D, 0x6B, 0x78, 0x2F, 0x30, 0x34, 0x28, 0xAB, 0x18, 0x29, 0xE };
ULONG	nt40[]  = { 0x51, 0x71, 0x80, 0x31, 0x32, 0x37, 0x2A, 0xB2, 0x19, 0x2C, 0xF };
// pointer to table in effect
PULONG	nttable;

// are we on nt 40?
BOOLEAN IsNT40 = FALSE;

// real registry routines
NTSTATUS (*RealRegOpenKey)( IN PHANDLE, IN OUT ACCESS_MASK, IN POBJECT_ATTRIBUTES );
NTSTATUS (*RealRegQueryKey)( IN HANDLE, IN KEY_INFORMATION_CLASS,
				OUT PVOID, IN ULONG, OUT PULONG );
NTSTATUS (*RealRegQueryValueKey)( IN HANDLE, IN PUNICODE_STRING, 
				IN KEY_VALUE_INFORMATION_CLASS,
				OUT PVOID, IN ULONG, OUT PULONG );
NTSTATUS (*RealRegEnumerateValueKey)( IN HANDLE, IN ULONG,  
				IN KEY_VALUE_INFORMATION_CLASS,
				OUT PVOID, IN ULONG, OUT PULONG );
NTSTATUS (*RealRegEnumerateKey)( IN HANDLE, IN ULONG,
				IN KEY_INFORMATION_CLASS,
				OUT PVOID, IN ULONG, OUT PULONG );
NTSTATUS (*RealRegSetValueKey)( IN HANDLE KeyHandle, IN PUNICODE_STRING ValueName,
				IN ULONG TitleIndex, IN ULONG Type, 
				IN PVOID Data, IN ULONG DataSize );
NTSTATUS (*RealRegCreateKey)( OUT PHANDLE, IN ACCESS_MASK,
				IN POBJECT_ATTRIBUTES , IN ULONG,
				IN PUNICODE_STRING, IN ULONG, OUT PULONG );
NTSTATUS (*RealRegDeleteValueKey)( IN HANDLE, IN PUNICODE_STRING );
NTSTATUS (*RealRegCloseKey)( IN HANDLE );
NTSTATUS (*RealRegDeleteKey)( IN HANDLE );
NTSTATUS (*RealRegFlushKey)( IN HANDLE );

// lenghs of rootkeys (filled in at init)
ROOTKEY RootKey[NUMROOTKEYS] = {
	{ "\\\\REGISTRY\\USER", "USERS", 0 },
	{ "\\\\REGISTRY\\MACHINE\\SYSTEM\\CURRENTCONTROLSET\\HARDWARE PROFILES\\CURRENT", 
		"CONFIG", 0 },
	{ "\\\\REGISTRY\\MACHINE\\SOFTWARE\\CLASSES", "ROOT", 0 },
	{ "\\\\REGISTRY\\MACHINE", "MACHINE", 0 }
};

// hash table for keeping names around. This is necessary because 
// at any time the name information in the registry bjects that we
// see can be deallocated and reused. If we want to print accurate
// names, we need to keep them around ourselves.
PHASH_ENTRY             HashTable[NUMHASH];

// list of freed entries (to save some allocation calls)
PHASH_ENTRY             FreeEntries = NULL;

// mutex for hash table accesses
MUTEX_TYPE              HashMutex;

// make sure we don't unload when there are outstanding registry calls
MUTEX_TYPE              RegMutex;

// Data structure for storing messages we generate
PSTORE_BUF              Store           = NULL;
ULONG                   Sequence        = 0;
MUTEX_TYPE              StoreMutex;

// maximum amount of data we will grab for buffered unread data
ULONG                   NumStore        = 0;
ULONG                   MaxStore        = 1000000/MAX_STORE;


//======================================================================
//
// NtQueryObject
//
// Since NT 3.51 does not export this call, we have to implement it
// ourselves. In addition, the system call number is different under 
// NT 3.51.
//
//======================================================================
NTSTATUS NtQueryObject( PHANDLE Handle, ULONG QueryType,  POBQUERYTYPE objtype, 
	ULONG Len, PULONG RetLen )
{
	NTSTATUS        ntstatus;
	ULONG		callno = IsNT40 ? 0x74 : 0x6D;

	_asm {
		mov     eax, callno
		lea     edx, Handle
		int     2eh     
		mov     ntstatus, eax
	}
	return ntstatus;
}


//======================================================================
//                   B U F F E R  R O U T I N E S 
//======================================================================

//----------------------------------------------------------------------
//
// RegmonFreeStore
//
// Frees all the data output buffers that we have currently allocated.
//
//----------------------------------------------------------------------
VOID RegmonFreeStore()
{
	PSTORE_BUF      prev;
    
	while( Store ) {
		prev = Store->Next;
		ExFreePool( Store );
		Store = prev;
	}
}       


//----------------------------------------------------------------------
//
// RegmonNewStore
//
// Called when the current buffer has filled up. This moves us to the
// pre-allocated buffer and then allocates another buffer.
//
//----------------------------------------------------------------------
void RegmonNewStore( void )
{
	PSTORE_BUF prev = Store, newstore;

	 // if we have maxed out or haven't accessed the current store
	// just return
	if( MaxStore == NumStore ) {
		Store->Len = 0;
		return; 
	}

	// see if we can re-use a store
	if( !Store->Len ) 
		return;

    	// move to the next buffer and allocate another one
    	newstore = ExAllocatePool( NonPagedPool, sizeof(*Store) );
    	if( newstore ) { 
    		Store 	= newstore;
    		Store->Len  = 0;
    		Store->Next = prev;
		NumStore++;
    	} else 
		Store->Len = 0;
}


//----------------------------------------------------------------------
//
// RegmonOldestStore
//
// Goes through the linked list of storage buffers and returns the 
// oldest one.
//
//----------------------------------------------------------------------
PSTORE_BUF RegmonOldestStore( void )
{
	PSTORE_BUF  ptr = Store, prev = NULL;

	while ( ptr->Next )
		ptr = (prev = ptr)->Next;
	if ( prev )
		prev->Next = NULL;    
	return ptr;
}


//----------------------------------------------------------------------
//
// RegmonResetStore
//
// When a GUI is no longer communicating with us, but we can't unload,
// we reset the storage buffers.
//
//----------------------------------------------------------------------
VOID RegmonResetStore()
{
   PSTORE_BUF  nextstore, curstore;

   curstore = Store->Next;
   while( curstore ) {
        nextstore = curstore->Next;
	ExFreePool( curstore );
	curstore = nextstore;
   }
   Store->Len = 0;
   Store->Next = NULL;
}


//----------------------------------------------------------------------
//
// UpdateStore
//
// Add a new string to Store, if it fits.
//
//----------------------------------------------------------------------
void UpdateStore( ULONG seq, const char * format, ... )
{       
	PENTRY          Entry;
	int             len;
	va_list         arg_ptr;

	// Let Mark see debug output we absconded.  This is x86 specific.
#define A (&format)
	DbgPrint(( (char *)format, A[1], A[2], A[3], A[4], A[5], A[6] ));
	DbgPrint(( "\n" ));
#undef A

	// only do this if a GUI is active
	if( !GUIActive ) return;

	MUTEX_P( StoreMutex );
    
	// See if its time to switch to extra buffer
	if ( Store->Len > MAX_STORE-350 )  
		RegmonNewStore();

	if ( Store->Len <= MAX_STORE-350 )  { 

		/* Set our sequence number */
		Entry = (void *)(Store->Data+Store->Len);
		Entry->seq = seq;

		va_start( arg_ptr, format );
		len = vsprintf( Entry->text, format, arg_ptr );
		va_end( arg_ptr );

		Store->Len += sizeof(Entry->seq)+len+1; // +1 for null char
	}
	MUTEX_V( StoreMutex );
}


//----------------------------------------------------------------------
// 
// strncatZ
//
// Appends a string to another and attaches a null.
//
//----------------------------------------------------------------------
PCHAR strncatZ( PCHAR dest, PCHAR source, int length )
{       
	int     origlen = strlen(dest);

	strncpy( dest+origlen, source, length );
	dest[ origlen+length ] = 0;
	return(dest);
}

//======================================================================
//                   H A S H  R O U T I N E S 
//======================================================================

//----------------------------------------------------------------------
//
// RegmonHashCleanup
//
// Called when we are unloading to free any memory that we have 
// in our possession.
//
//----------------------------------------------------------------------
VOID RegmonHashCleanup()
{
	PHASH_ENTRY             hashEntry, nextEntry;
	ULONG           i;
  
	// first free the hash table entries
	for( i = 0; i < NUMHASH; i++ ) {
		hashEntry = HashTable[i];
		while( hashEntry ) {
			nextEntry = hashEntry->Next;
			ExFreePool( hashEntry->FullPathName );
			ExFreePool( hashEntry );
			hashEntry = nextEntry;
		}
	}

	// now, free structures on our free list
	while( FreeEntries ) {
		nextEntry = FreeEntries->Next;
		ExFreePool( FreeEntries );
		FreeEntries = nextEntry;
	}
}

//----------------------------------------------------------------------
//
// RegmonStoreHash
//
// Stores the key and associated fullpath in the hash table.
//
//----------------------------------------------------------------------
VOID RegmonStoreHash( POBJECT object, PCHAR fullname )
{
	PHASH_ENTRY     newEntry;

	MUTEX_P( HashMutex );
	if( FreeEntries ) {
		newEntry    = FreeEntries;
		FreeEntries = FreeEntries->Next;
	} else 
		newEntry = ExAllocatePool( NonPagedPool, sizeof(HASH_ENTRY) );

	newEntry->Object                = object;
	newEntry->FullPathName          = ExAllocatePool( NonPagedPool, strlen(fullname)+1 );
	newEntry->Next                  = HashTable[ HASHOBJECT( object) ];
	HashTable[ HASHOBJECT(object) ] = newEntry;     
	strcpy( newEntry->FullPathName, fullname );
	MUTEX_V( HashMutex );
}


//----------------------------------------------------------------------
//
// RegmonFreeHashEntry
//
// When we see a file close, we can free the string we had associated
// with the fileobject being closed since we know it won't be used
// again.
//
//----------------------------------------------------------------------
VOID RegmonFreeHashEntry( POBJECT object )
{
	PHASH_ENTRY             hashEntry, prevEntry;

	MUTEX_P( HashMutex );

	// look-up the entry
	hashEntry = HashTable[ HASHOBJECT( object ) ];
	prevEntry = NULL;
	while( hashEntry && hashEntry->Object != object ) {
		prevEntry = hashEntry;
		hashEntry = hashEntry->Next;
	}
  
	// if we fall off, just return
	if( !hashEntry ) {
		MUTEX_V( HashMutex );
		return;
	}

	// remove it from the hash list 
	if( prevEntry ) 
		prevEntry->Next = hashEntry->Next;
	else 
		HashTable[ HASHOBJECT( object )] = hashEntry->Next;

	// free the memory associated with it
	ExFreePool( hashEntry->FullPathName );

	// put it on our free list
	hashEntry->Next         = FreeEntries;
	FreeEntries             = hashEntry;

	MUTEX_V( HashMutex );
}

//======================================================================
//  R E G I S T R Y  P A R A M E T E R  S U P P O R T  R O U T I N E S
//======================================================================

//----------------------------------------------------------------------
//
// ConverToUpper
//
// Obvious.
//
//----------------------------------------------------------------------
VOID ConvertToUpper( PCHAR Dest, PCHAR Source, ULONG Len )
{
	ULONG   i;
	
	for( i = 0; i < Len; i++ ) {
		if( Source[i] >= 'a' && Source[i] <= 'z' )
			Dest[i] = Source[i] - 'a' + 'A';
		else
			Dest[i] = Source[i];
	}
}


//----------------------------------------------------------------------
//
// GetPointer
//
// Translates a handle to an object pointer.
//
//----------------------------------------------------------------------
POBJECT GetPointer( HANDLE handle )
{
	POBJECT         pKey;

	// ignore null handles
	if( !handle ) return NULL;

	// get the pointer the handle refers to
	if( ObReferenceObjectByHandle( handle, 0, NULL, KernelMode, &pKey, NULL ) !=
			STATUS_SUCCESS ) {
		DbgPrint(("Error %x getting key pointer\n"));
		pKey = NULL;
	} 
	return pKey;
}


//----------------------------------------------------------------------
//
// ReleasePointer
//
// Dereferences the object.
//
//----------------------------------------------------------------------
VOID ReleasePointer( POBJECT object )
{
	if( object ) ObDereferenceObject( object );
}


//----------------------------------------------------------------------
//
// AppendKeyInformation
//
// Appends key enumerate and query information to the output buffer.
//
//----------------------------------------------------------------------
VOID AppendKeyInformation( IN KEY_INFORMATION_CLASS KeyInformationClass,
	IN PVOID KeyInformation, PCHAR Buffer )
{
	PKEY_BASIC_INFORMATION  pbasicinfo;
	PKEY_FULL_INFORMATION   pfullinfo;
	PKEY_NODE_INFORMATION   pnodeinfo;
	UNICODE_STRING          ukeyname;       
	ANSI_STRING             akeyname;

	switch( KeyInformationClass ) {

	case KeyBasicInformation:
		pbasicinfo = (PKEY_BASIC_INFORMATION) KeyInformation;
		ukeyname.Length = (USHORT) pbasicinfo->NameLength;
		ukeyname.MaximumLength = (USHORT) pbasicinfo->NameLength;
		ukeyname.Buffer = pbasicinfo->Name;
		RtlUnicodeStringToAnsiString( &akeyname, &ukeyname, TRUE );
		sprintf( Buffer, "Name: %s", akeyname.Buffer );
		RtlFreeAnsiString( &akeyname );
		break;
	case KeyFullInformation:
		pfullinfo = (PKEY_FULL_INFORMATION) KeyInformation;
		sprintf( Buffer, "Subkeys = %d", pfullinfo->SubKeys );
		break;  
		
	case KeyNodeInformation:
		pnodeinfo = (PKEY_NODE_INFORMATION) KeyInformation;
		ukeyname.Length = (USHORT) pnodeinfo->ClassLength;
		ukeyname.MaximumLength = (USHORT) pnodeinfo->ClassLength;
		ukeyname.Buffer = (PWCHAR) ((PCHAR) pnodeinfo + 
			pnodeinfo->ClassOffset);
		RtlUnicodeStringToAnsiString( &akeyname, &ukeyname, TRUE );
		sprintf( Buffer, "Name: %s", akeyname.Buffer );
		RtlFreeAnsiString( &akeyname );
		break;
	default:
		sprintf( Buffer, "Unknown Info Class" );
		break;
	}
}


//----------------------------------------------------------------------
//
// AppendRegValueType
//
// Returns the string form of an registry value type.
//
//----------------------------------------------------------------------
VOID AppendRegValueType( ULONG Type, PCHAR Buffer )
{
	CHAR            tmp[32];

	switch( Type ) {
	case REG_BINARY:
		strcat( Buffer, "BINARY" );
		break;
	case REG_DWORD_LITTLE_ENDIAN:
		strcat( Buffer, "DWORD_LITTLE_END" );
		break;
	case REG_DWORD_BIG_ENDIAN:
		strcat( Buffer, "DWORD_BIG_END" );
		break;
	case REG_EXPAND_SZ:
		strcat( Buffer, "EXPAND_SZ" );
		break;
	case REG_LINK:
		strcat( Buffer, "LINK" );
		break;
	case REG_MULTI_SZ:
		strcat( Buffer, "MULTI_SZ" );
		break;
	case REG_NONE:
		strcat( Buffer, "NONE" );
		break;
	case REG_SZ:
		strcat( Buffer, "SZ" );
		break;
	case REG_RESOURCE_LIST:
		strcat( Buffer, "RESOURCE_LIST" );
		break;
	case REG_RESOURCE_REQUIREMENTS_LIST:
		strcat( Buffer, "REQ_LIST" );
		break;
	case REG_FULL_RESOURCE_DESCRIPTOR:
		strcat( Buffer, "DESCRIPTOR" );
		break;
	default:
		sprintf( tmp, "UNKNOWN TYPE: %d", Type );
		strcat( Buffer, tmp );
		break;
	}
}


//----------------------------------------------------------------------
//
// AppendRegValueData
//
// We expand certain registry types to provide more information.
//
//----------------------------------------------------------------------
VOID AppendRegValueData( IN ULONG Type, IN PVOID Data, IN ULONG Length, 
		IN OUT PCHAR Buffer )
{
	PWCHAR                  pstring;
	PULONG                  pulong;
	PUCHAR                  pbinary;
	CHAR                    tmp[16];
	UNICODE_STRING          ukeyname;       
	ANSI_STRING             akeyname;
	int                     len, i;

	switch( Type ) {
	case REG_SZ:    
	case REG_EXPAND_SZ:
	case REG_MULTI_SZ:
		pstring = (PWCHAR) Data;
		ukeyname.Length = (USHORT) Length;
		ukeyname.MaximumLength = (USHORT) Length;
		ukeyname.Buffer = pstring;
		RtlUnicodeStringToAnsiString( &akeyname, 
			&ukeyname, TRUE );      
		strcat( Buffer, "\"");
		strcat( Buffer+1, akeyname.Buffer );
		strcat( Buffer, "\"");
		RtlFreeAnsiString( &akeyname );
		break;
	case REG_DWORD:
		pulong = (PULONG) Data;
		sprintf( tmp, "0x%X", *pulong );
		strcat( Buffer, tmp );
		break;
	case REG_BINARY:
	case REG_RESOURCE_LIST:
	case REG_FULL_RESOURCE_DESCRIPTOR:
	case REG_RESOURCE_REQUIREMENTS_LIST:
		pbinary = (PCHAR) Data;
		if( Length > 8 ) len = 8;
		else len = Length;
		for( i = 0; i < len; i++ ) {
			sprintf( tmp, "%02X ", (UCHAR) pbinary[i]);
			strcat( Buffer, tmp );
		}
		if( Length > 8) strcat( Buffer, "...");
		break;
	default:
		AppendRegValueType( Type, Buffer );
		break;
	}
}


//----------------------------------------------------------------------
//
// AppendValueInformation
//
// Appends value enumerate and query information to the output buffer.
//
//----------------------------------------------------------------------
VOID AppendValueInformation( IN KEY_VALUE_INFORMATION_CLASS KeyValueInformationClass,
	IN PVOID KeyValueInformation, PCHAR Buffer, PCHAR ValueName )
{
	PKEY_VALUE_BASIC_INFORMATION    pbasicinfo;
	PKEY_VALUE_FULL_INFORMATION     pfullinfo;
	PKEY_VALUE_PARTIAL_INFORMATION  ppartinfo;
	UNICODE_STRING                  ukeyname;       
	ANSI_STRING                     akeyname;

	switch( KeyValueInformationClass ) {

	case KeyValueBasicInformation:
		pbasicinfo = (PKEY_VALUE_BASIC_INFORMATION)
				KeyValueInformation;
		sprintf( Buffer, "Type: ");
		AppendRegValueType( pbasicinfo->Type, Buffer );
		strcat( Buffer, " Name: " );
		ukeyname.Length = (USHORT) pbasicinfo->NameLength;
		ukeyname.MaximumLength = (USHORT) pbasicinfo->NameLength;
		ukeyname.Buffer = pbasicinfo->Name;
		RtlUnicodeStringToAnsiString( &akeyname, &ukeyname, TRUE );
		strcat( Buffer, akeyname.Buffer );
		if( ValueName ) strcpy( ValueName, akeyname.Buffer );
		RtlFreeAnsiString( &akeyname );                 
		break;

	case KeyValueFullInformation:   
		pfullinfo = (PKEY_VALUE_FULL_INFORMATION) 
				KeyValueInformation;
		AppendRegValueData( pfullinfo->Type, 
			(PVOID) ((PCHAR) pfullinfo + pfullinfo->DataOffset), 
			pfullinfo->DataLength, Buffer );
		if( ValueName ) {
			ukeyname.Length = (USHORT) pfullinfo->NameLength;
			ukeyname.MaximumLength = (USHORT) pfullinfo->NameLength;
			ukeyname.Buffer = pfullinfo->Name;
			RtlUnicodeStringToAnsiString( &akeyname, &ukeyname, TRUE );
			strcpy( ValueName, akeyname.Buffer );
			RtlFreeAnsiString( &akeyname ); 
		}
		break;

	case KeyValuePartialInformation:
		ppartinfo = (PKEY_VALUE_PARTIAL_INFORMATION)
				KeyValueInformation;
		AppendRegValueData( ppartinfo->Type, 
			(PVOID) ppartinfo->Data, 
			ppartinfo->DataLength, Buffer );
		break;

	default:
		sprintf( Buffer, "Unknown Info Class" );
		break;
	}
}

//----------------------------------------------------------------------
//
// ErrorString
//
// Returns the string form of an error code.
//
//----------------------------------------------------------------------
PCHAR ErrorString( NTSTATUS retval )
{
	switch( retval ) {
	case STATUS_SUCCESS:
		return "SUCCESS";
	case STATUS_KEY_DELETED:
		return "KEYDELETED";
	case STATUS_REGISTRY_IO_FAILED:
		return "IOFAILED";
	case STATUS_REGISTRY_CORRUPT:
		return "CORRUPT";
	case STATUS_NO_MEMORY:
		return "OUTOFMEM";
	case STATUS_ACCESS_DENIED:
		return "ACCDENIED";
	case STATUS_NO_MORE_ENTRIES:
		return "NOMORE";
	case STATUS_OBJECT_NAME_NOT_FOUND:
		return "NOTFOUND";
	case STATUS_BUFFER_OVERFLOW:
		return "BUFOVRFLOW";
	case STATUS_OBJECT_PATH_SYNTAX_BAD:
		return "SYNTAXERR";
	default:
		sprintf(errstring, "%x", retval );
		return errstring;
	}
}

//----------------------------------------------------------------------
//
// GetFullName
//
// Returns the full pathname of a key, if we can obtain one, else
// returns a handle.
//
//----------------------------------------------------------------------
void GetFullName( HANDLE hKey, PUNICODE_STRING lpszSubKeyVal, 
		PCHAR fullname )
{
	PHASH_ENTRY             hashEntry;
	POBJECT                 pKey = NULL;
	CHAR                    tmpkey[16];
	ANSI_STRING             keyname;
	CHAR                    tmpname[1024];
	CHAR                    cmpname[256];
	PCHAR                   nameptr;
	int                     i;

	// first, translate the hkey into a pointer
	fullname[0] = 0;
	tmpname[0] = 0;

	// is it a valid handle?
	if( pKey = GetPointer( hKey )) {

		// see if we find the key in the hash table
		ReleasePointer( pKey );
		MUTEX_P( HashMutex );
		hashEntry = HashTable[ HASHOBJECT( pKey ) ];
		while( hashEntry && hashEntry->Object != pKey )
		hashEntry = hashEntry->Next;
		MUTEX_V( HashMutex );
		if( hashEntry )
			strcpy( tmpname, hashEntry->FullPathName );
		else {
			// we will only get here if key was created before we loaded
			if( pKey ) {
				sprintf( tmpkey, "0x%X", pKey );                                        
				strcat(tmpname, tmpkey );
			}
		}
	}

	// append subkey and value, if they are there
	if( lpszSubKeyVal ) {
		RtlUnicodeStringToAnsiString( &keyname, lpszSubKeyVal, TRUE );
		if( keyname.Buffer[0] ) {
			strcat( tmpname, "\\" );
			strcat( tmpname, keyname.Buffer );
		}
		RtlFreeAnsiString( &keyname );
	}

	// now, see if we can translate a root key name
	for( i = 0; i < NUMROOTKEYS; i++ ) {
		ConvertToUpper( cmpname, tmpname, RootKey[i].RootNameLen );
		if( !strncmp( cmpname, RootKey[i].RootName, 
				RootKey[i].RootNameLen )) {
			nameptr = tmpname + RootKey[i].RootNameLen;
			strcpy( fullname, RootKey[i].RootShort );
			strcat( fullname, nameptr );
			return;
		}
	}

	// no translation
	strcpy( fullname, tmpname );
}


//----------------------------------------------------------------------
//
// GetProcess
//
// Uses undocumented data structure offsets to obtain the name of the
// currently executing process.
//
//----------------------------------------------------------------------
PCHAR GetProcess( PCHAR Name )
{
    PEPROCESS	  	curproc;
    char                *nameptr;

    curproc = PsGetCurrentProcess();
    nameptr   = (char *) ((ULONG)curproc + NAMEOFFSET);
    strncpy( Name, nameptr, 16 );
    return Name;
}

//======================================================================
//                    H O O K  R O U T I N E S
//======================================================================

//----------------------------------------------------------------------
//
// HookRegOpenKey
//
//----------------------------------------------------------------------
NTSTATUS HookRegOpenKey( IN OUT PHANDLE pHandle, IN ACCESS_MASK ReqAccess, 
	IN POBJECT_ATTRIBUTES pOpenInfo )
{
	NTSTATUS                ntstatus;
	POBJECT                 regobj;
	CHAR                    fullname[1024], data[32], name[20];

	GetFullName( pOpenInfo->RootDirectory, pOpenInfo->ObjectName, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegOpenKey( pHandle, ReqAccess, pOpenInfo );
	MUTEX_V( RegMutex );
	DbgPrint(("RegOpenKey: %s => %x, %x\n", fullname, *pHandle, ntstatus ));
	data[0] = 0;
	if( NT_SUCCESS( ntstatus )) {   
		regobj = GetPointer( *pHandle );
		RegmonFreeHashEntry( regobj );
		RegmonStoreHash( regobj, fullname );
		sprintf(data,"Key: 0x%X", regobj );
		ReleasePointer( regobj );
	}
	UpdateStore( Sequence++, "%s\tOpenKey\t%s\t%s\t%s", GetProcess(name),
		fullname, data, ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegCreateKey
//
//----------------------------------------------------------------------
NTSTATUS HookRegCreateKey( OUT PHANDLE pHandle, IN ACCESS_MASK ReqAccess,
	IN POBJECT_ATTRIBUTES pOpenInfo, IN ULONG TitleIndex,
	IN PUNICODE_STRING Class, IN ULONG CreateOptions, OUT PULONG Disposition )
{
	NTSTATUS                ntstatus;
	POBJECT                 regobj;
	CHAR                    fullname[1024], data[32], name[20];

	GetFullName( pOpenInfo->RootDirectory, pOpenInfo->ObjectName, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegCreateKey( pHandle, ReqAccess, pOpenInfo, TitleIndex,
			Class, CreateOptions, Disposition );
	MUTEX_V( RegMutex );
	DbgPrint(("RegCreateKey: %s => %x, %x\n", fullname, *pHandle, ntstatus ));
	data[0] = 0;
	if( NT_SUCCESS( ntstatus )) {   
		regobj = GetPointer( *pHandle );
		RegmonFreeHashEntry( regobj );
		RegmonStoreHash( regobj, fullname );
		sprintf(data,"Key: 0x%X", regobj );
		ReleasePointer( regobj );
	}
	UpdateStore( Sequence++, "%s\tCreateKey\t%s\t%s\t%s", GetProcess(name),
		fullname, data,	ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegCloseKey
//
// This is actually a hook for NtClose which is used for closing any
// object. Therefore, we must ensure that we are seeing a close for 
// a registry object that we know about.
//
//----------------------------------------------------------------------
NTSTATUS HookRegCloseKey( IN HANDLE Handle )
{
	NTSTATUS                ntstatus;
	POBJECT                 regobj;
	CHAR                    queryinfo[1024];
	CHAR                    fullname[1024], data[256], name[20];
	POBQUERYTYPE            objtype;        
	ANSI_STRING             atype;
	ULONG                   retlen;
	BOOLEAN                 iskey = FALSE;

	// see if its a registry key being closed
	objtype = (POBQUERYTYPE) queryinfo;
	retlen = 0;
	ntstatus = NtQueryObject( Handle, QUERYOBJTYPE,  objtype, 1024, &retlen );
	if( NT_SUCCESS( ntstatus )) {
		RtlUnicodeStringToAnsiString( &atype, &objtype->Name, TRUE );
		if( !strcmp( "Key", atype.Buffer )) {
			iskey = TRUE;
			GetFullName( Handle, NULL, fullname );

			// get the pointer for later
			regobj = GetPointer( Handle );
			ReleasePointer( regobj );
		}
		RtlFreeAnsiString( &atype );
	}
	MUTEX_P( RegMutex );
	ntstatus = RealRegCloseKey( Handle );
	MUTEX_V( RegMutex );
	if( iskey ) {
		DbgPrint(("RegCloseKey: %s => %x, %x\n", fullname, Handle, ntstatus ));
		data[0] = 0;
		if( NT_SUCCESS( ntstatus )) {
			if( regobj ) RegmonFreeHashEntry( regobj );
			sprintf(data,"Key: 0x%X", regobj );
			UpdateStore( Sequence++, "%s\tCloseKey\t%s\t%s\t%s", 
				GetProcess(name), fullname,      
				data, ErrorString( ntstatus ));
		}
	}
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegFlushKey
//
//----------------------------------------------------------------------
NTSTATUS HookRegFlushKey( IN HANDLE Handle )
{
	NTSTATUS                ntstatus;
	CHAR                    fullname[1024], name[20];
	POBJECT                 regobj;

	GetFullName( Handle, NULL, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegFlushKey( Handle );
	MUTEX_V( RegMutex );
	DbgPrint(("RegFlushKey: %s => 0x%X\n", fullname, ntstatus ));
	regobj = GetPointer( Handle );
	ReleasePointer( regobj );
	UpdateStore( Sequence++, "%s\tFlushKey\t%s\tKey: 0x%X\t%s", 
		GetProcess( name ), fullname,
		regobj, ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegDeleteKey
//
// Once we've deleted a key, we can remove its reference in the hash 
// table.
//
//----------------------------------------------------------------------
NTSTATUS HookRegDeleteKey( IN HANDLE Handle )
{
	NTSTATUS                ntstatus;
	POBJECT                 regobj;
	CHAR                    fullname[1024], name[20];

	GetFullName( Handle, NULL, fullname );
	regobj = GetPointer( Handle );
	ReleasePointer( regobj );
	MUTEX_P( RegMutex );
	ntstatus = RealRegDeleteKey( Handle );
	MUTEX_V( RegMutex );
	DbgPrint(("RegDeleteKey: %s => 0x%X\n", fullname, ntstatus ));
	UpdateStore( Sequence++, "%s\tDeleteKey\t%s\tKey: 0x%X\t%s", 
		GetProcess( name ), fullname, 
		regobj, ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegDeleteValueKey
//
//----------------------------------------------------------------------
NTSTATUS HookRegDeleteValueKey( IN HANDLE Handle, PUNICODE_STRING Name )
{
	NTSTATUS                ntstatus;
	CHAR                    fullname[1024], name[20];

	GetFullName( Handle, Name, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegDeleteValueKey( Handle, Name );
	MUTEX_V( RegMutex );
	DbgPrint(("RegDeleteValueKey: %s => %x\n", fullname, ntstatus ));
	UpdateStore( Sequence++, "%s\tDeleteValueKey\t%s\t\t%s", 
		GetProcess( name ), fullname,
		ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegSetValueKey
//
//---------------------------------------------------------------------- 
NTSTATUS HookRegSetValueKey( IN HANDLE KeyHandle, IN PUNICODE_STRING ValueName,
	IN ULONG TitleIndex, IN ULONG Type, IN PVOID Data, IN ULONG DataSize )
{
	NTSTATUS                ntstatus;
	CHAR                    fullname[1024], data[256], name[20];

	GetFullName( KeyHandle, ValueName, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegSetValueKey( KeyHandle, ValueName, TitleIndex,
			Type, Data, DataSize );
	MUTEX_V( RegMutex );
	data[0] = 0;
	if( NT_SUCCESS( ntstatus )) 
		AppendRegValueData( Type, Data, DataSize, data );
	DbgPrint(("SetValue: %s (%s)\n", fullname, data ));
	UpdateStore( Sequence++, "%s\tSetValue\t%s\t%s\t%s", 
		GetProcess( name ), fullname, data,
		ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegEnumerateKey
//
// This is a documented Zw-class function. 
//
//----------------------------------------------------------------------
NTSTATUS HookRegEnumerateKey( IN HANDLE KeyHandle, IN ULONG Index,
	IN KEY_INFORMATION_CLASS KeyInformationClass,
	OUT PVOID KeyInformation, IN ULONG Length, OUT PULONG pResultLength )
{
	NTSTATUS                ntstatus;
	CHAR                    fullname[1024], data[256], name[20];

	GetFullName( KeyHandle, NULL, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegEnumerateKey( KeyHandle, Index, KeyInformationClass,
				KeyInformation, Length, pResultLength );
	MUTEX_V( RegMutex );
	data[0] = 0;
	if( NT_SUCCESS( ntstatus )) 
		AppendKeyInformation( KeyInformationClass, KeyInformation, data );

	DbgPrint(("EnumerateKey: %s (%s) => %x\n", fullname, data, ntstatus ));
	UpdateStore( Sequence++, "%s\tEnumerateKey\t%s\t%s\t%s", 
		GetProcess( name ), fullname, data,
		ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegQueryKey
//
// This is a documented Zw-class function. This is *rarely* called.
//
//----------------------------------------------------------------------
NTSTATUS HookRegQueryKey( IN HANDLE  KeyHandle, 
	IN KEY_INFORMATION_CLASS  KeyInformationClass,
	OUT PVOID  KeyInformation, IN ULONG  Length, 
	OUT PULONG  pResultLength )
{
	NTSTATUS                ntstatus;
	CHAR                    fullname[1024], data[256], name[20];

	GetFullName( KeyHandle, NULL, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegQueryKey( KeyHandle, KeyInformationClass,
			KeyInformation, Length, pResultLength );
	MUTEX_V( RegMutex );
	// print out different stuff depending on type of info asked for
	data[0] = 0;
	if( NT_SUCCESS( ntstatus )) 
		AppendKeyInformation( KeyInformationClass, KeyInformation, data );

	DbgPrint(("QueryKey: %s (%s) => %x\n", fullname, data, ntstatus ));
	UpdateStore( Sequence++, "%s\tQueryKey\t%s\t%s\t%s", 
		GetProcess( name ), fullname, data,
		ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegEnumerateValueKey
//
// This is a documented Zw-class function.
//
//----------------------------------------------------------------------
NTSTATUS HookRegEnumerateValueKey( IN HANDLE KeyHandle, IN ULONG Index,
	IN KEY_VALUE_INFORMATION_CLASS KeyValueInformationClass,
	OUT PVOID KeyValueInformation, IN ULONG Length,
	OUT PULONG  pResultLength )
{
	NTSTATUS                ntstatus;
	CHAR                    fullname[1024], data[256], valuename[256], name[20];

	GetFullName( KeyHandle, NULL, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegEnumerateValueKey( KeyHandle, Index,
			KeyValueInformationClass,
			KeyValueInformation, Length, 
			pResultLength );
	MUTEX_V( RegMutex );
	data[0] = 0;
	if( NT_SUCCESS( ntstatus )) {
		AppendValueInformation( KeyValueInformationClass, 
			KeyValueInformation, data, valuename );
		strcat( fullname, "\\" );
		strcat( fullname, valuename );
	}
	DbgPrint(("EnumerateValue: %s (%s) =>%x\n", fullname, data, ntstatus ));
	UpdateStore( Sequence++, "%s\tEnumerateValue\t%s\t%s\t%s", 
		GetProcess( name ), fullname, data,
		ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegQueryValueKey
//
// This is a documented Zw-class function.
//
//----------------------------------------------------------------------
NTSTATUS HookRegQueryValueKey( IN HANDLE KeyHandle,
	IN PUNICODE_STRING ValueName,
	IN KEY_VALUE_INFORMATION_CLASS KeyValueInformationClass,
	OUT PVOID KeyValueInformation, IN ULONG Length,
	OUT PULONG  pResultLength )
{
	NTSTATUS                ntstatus;
	CHAR                    fullname[1024], data[256], name[20];

	GetFullName( KeyHandle, ValueName, fullname );
	MUTEX_P( RegMutex );
	ntstatus = RealRegQueryValueKey( KeyHandle, ValueName,
			KeyValueInformationClass,
			KeyValueInformation, Length, 
			pResultLength );
	MUTEX_V( RegMutex );
	data[0] = 0;
	if( NT_SUCCESS( ntstatus )) 
		AppendValueInformation( KeyValueInformationClass, 
			KeyValueInformation, data, FALSE );

	DbgPrint(("QueryValue: %s (%s) =>%x\n", fullname, data, ntstatus ));
	UpdateStore( Sequence++, "%s\tQueryValue\t%s\t%s\t%s", 
		GetProcess( name ), fullname, data,
		ErrorString( ntstatus ));
	return ntstatus;
}


//----------------------------------------------------------------------
//
// HookRegistry
//
// Replaces entries in the system service table with pointers to
// our own hook routines. We save off the real routine addresses.
//
//----------------------------------------------------------------------
VOID HookRegistry( void )
{
	// hook everything
	RealRegOpenKey = SYSCALL( REGOPENKEY );
	SYSCALL( REGOPENKEY ) = (PVOID) HookRegOpenKey;

	RealRegQueryKey = SYSCALL( REGQUERYKEY );
	SYSCALL( REGQUERYKEY ) = (PVOID) HookRegQueryKey;

	RealRegQueryValueKey = SYSCALL( REGQUERYVALUEKEY );
	SYSCALL( REGQUERYVALUEKEY ) = (PVOID) HookRegQueryValueKey;

	RealRegEnumerateValueKey = SYSCALL( REGENUMERATEVALUEKEY );
	SYSCALL( REGENUMERATEVALUEKEY ) = 
			(PVOID) HookRegEnumerateValueKey;

	RealRegEnumerateKey = SYSCALL( REGENUMERATEKEY );
	SYSCALL( REGENUMERATEKEY ) = (PVOID) HookRegEnumerateKey;

	RealRegDeleteKey = SYSCALL( REGDELETEKEY );
	SYSCALL( REGDELETEKEY ) = (PVOID) HookRegDeleteKey;

	RealRegFlushKey = SYSCALL( REGFLUSHKEY );
	SYSCALL( REGFLUSHKEY ) = (PVOID) HookRegFlushKey;

	RealRegSetValueKey = SYSCALL( REGSETVALUEKEY );
	SYSCALL( REGSETVALUEKEY ) = (PVOID) HookRegSetValueKey;

	RealRegCreateKey = SYSCALL( REGCREATEKEY );
	SYSCALL( REGCREATEKEY ) = (PVOID) HookRegCreateKey;

	RealRegDeleteValueKey = SYSCALL( REGDELETEVALUEKEY );
	SYSCALL( REGDELETEVALUEKEY ) = (PVOID) HookRegDeleteValueKey;
	
	RealRegCloseKey = SYSCALL( REGCLOSEKEY );
	SYSCALL( REGCLOSEKEY ) = (PVOID) HookRegCloseKey;
	RegHooked = TRUE;
}


//----------------------------------------------------------------------
//
// UnhookRegistry
//
// Unhooks all registry routines by replacing the hook addresses in 
// the system service table with the real routine addresses that we
// saved off.
//
//----------------------------------------------------------------------
VOID UnhookRegistry( )
{
	// unhook everything
	MUTEX_P( RegMutex );
	SYSCALL( REGOPENKEY ) = (PVOID) RealRegOpenKey;
	SYSCALL( REGQUERYKEY ) = (PVOID) RealRegQueryKey;
	SYSCALL( REGQUERYVALUEKEY ) = (PVOID) RealRegQueryValueKey;
	SYSCALL( REGENUMERATEVALUEKEY ) = 
			(PVOID) RealRegEnumerateValueKey;
	SYSCALL( REGENUMERATEKEY ) = (PVOID) RealRegEnumerateKey;
	SYSCALL( REGCLOSEKEY ) = (PVOID) RealRegCloseKey;
	SYSCALL( REGFLUSHKEY ) = (PVOID) RealRegFlushKey;
	SYSCALL( REGDELETEKEY ) = (PVOID) RealRegDeleteKey;
	SYSCALL( REGSETVALUEKEY ) = (PVOID) RealRegSetValueKey;
	SYSCALL( REGCREATEKEY ) = (PVOID) RealRegCreateKey;
	SYSCALL( REGDELETEVALUEKEY ) = (PVOID) RealRegDeleteValueKey;
	RegHooked = FALSE;
	MUTEX_V( RegMutex );
}

//======================================================================
//         D E V I C E - D R I V E R  R O U T I N E S 
//======================================================================

//----------------------------------------------------------------------
//
// RegmonDeviceControl
//
//----------------------------------------------------------------------
BOOLEAN  RegmonDeviceControl( IN PFILE_OBJECT FileObject, IN BOOLEAN Wait,
		IN PVOID InputBuffer, IN ULONG InputBufferLength, 
		OUT PVOID OutputBuffer, IN ULONG OutputBufferLength, 
		IN ULONG IoControlCode, OUT PIO_STATUS_BLOCK IoStatus, 
		IN PDEVICE_OBJECT DeviceObject ) {
	BOOLEAN                 retval = FALSE;
	PSTORE_BUF              old;

	// Its a message from our GUI!
	IoStatus->Status      = STATUS_SUCCESS; // Assume success
	IoStatus->Information = 0;              // Assume nothing returned
	switch ( IoControlCode ) {
	case REGMON_hook:
		DbgPrint (("Regmon: hook\n"));
		HookRegistry();
		break;
	case REGMON_unhook:
		DbgPrint(("Regmon: unhook\n"));
		UnhookRegistry();
		break;
	case REGMON_zerostats:
		// Zero contents of buffer
		DbgPrint (("Regmon: zero stats\n"));
		MUTEX_P( StoreMutex );
		while ( Store->Next )  {
			// release next
			old = Store->Next;
			Store->Next = old->Next;
			MUTEX_P( StoreMutex );
			ExFreePool( old );
			MUTEX_V( StoreMutex );
			NumStore--;
		}
		Store->Len = 0;
		MUTEX_V( StoreMutex );
		break;
	case REGMON_getstats:
		// Copy buffer into user space.
		DbgPrint (("Regmon: get stats\n"));
		MUTEX_P( StoreMutex );
		if ( MAX_STORE > OutputBufferLength )  {
			// Don't do this; its not worth implementing the fix.
			MUTEX_V( StoreMutex );
			return FALSE;
		} else if ( Store->Len  ||  Store->Next ) {

			// Switch to a new store
			RegmonNewStore();

			// Fetch the oldest to give to user
			old = RegmonOldestStore();
			MUTEX_V( StoreMutex );

			// Copy it
			memcpy( OutputBuffer, old->Data, old->Len );

			// Return length of copied info
			IoStatus->Information = old->Len;

			// Deallocate buffer
			ExFreePool( old );
			NumStore--;
		} else {
			// no unread data
			MUTEX_V( StoreMutex );
			IoStatus->Information = 0;
		}
		break;
 
	default:
		DbgPrint (("Regmon: unknown IRP_MJ_DEVICE_CONTROL\n"));
		IoStatus->Status = 2;   // Bad request
		break;
	}
	return TRUE;
}


//----------------------------------------------------------------------
//
// RegmonDispatch
//
// In this routine we handle requests to our own device. The only 
// requests we care about handling explicitely are IOCTL commands that
// we will get from the GUI. We also expect to get Create and Close 
// commands when the GUI opens and closes communications with us.
//
//----------------------------------------------------------------------
NTSTATUS RegmonDispatch( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
	PIO_STACK_LOCATION      irpStack;
	PVOID                   inputBuffer;
	PVOID                   outputBuffer;
	ULONG                   inputBufferLength;
	ULONG                   outputBufferLength;
	ULONG                   ioControlCode;

	// go ahead and set the request up as successful
	Irp->IoStatus.Status      = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;

	// Get a pointer to the current location in the Irp. This is where
	//     the function codes and parameters are located.
	irpStack = IoGetCurrentIrpStackLocation (Irp);

	// Get the pointer to the input/output buffer and its length
	inputBuffer             = Irp->AssociatedIrp.SystemBuffer;
	inputBufferLength       = irpStack->Parameters.DeviceIoControl.InputBufferLength;
	outputBuffer            = Irp->AssociatedIrp.SystemBuffer;;
	outputBufferLength      = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
	ioControlCode           = irpStack->Parameters.DeviceIoControl.IoControlCode;

	switch (irpStack->MajorFunction) {
	case IRP_MJ_CREATE:
		DbgPrint(("Regmon: IRP_MJ_CREATE\n"));
		// fail multiple GUI connects at one time
		if( GUIActive )
			Irp->IoStatus.Status = 2;
		else
			Sequence = 0;
		GUIActive++;
		DbgPrint((" GUI Active: %d\n", GUIActive ));
		break;

	case IRP_MJ_CLOSE:
		DbgPrint(("Regmon: IRP_MJ_CLOSE\n"));
		GUIActive--;
		DbgPrint((" GUI closing: %d\n", GUIActive ));
		if( !GUIActive )
			RegmonResetStore();
		break;

	case IRP_MJ_DEVICE_CONTROL:
		DbgPrint (("Regmon: IRP_MJ_DEVICE_CONTROL\n"));

		// Its a request from the GUI
		RegmonDeviceControl( irpStack->FileObject, TRUE,
			     inputBuffer, inputBufferLength, 
			     outputBuffer, outputBufferLength,
			     ioControlCode, &Irp->IoStatus, DeviceObject );
		break;
	}
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;   
}


//----------------------------------------------------------------------
//
// RegmonUnload
//
// Our job is done - time to leave.
//
//----------------------------------------------------------------------
VOID RegmonUnload( IN PDRIVER_OBJECT DriverObject )
{
	WCHAR                   deviceLinkBuffer[]  = L"\\DosDevices\\Regmon";
	UNICODE_STRING          deviceLinkUnicodeString;

	DbgPrint(("Regmon.SYS: unloading\n"));

	// unhook 
	if( RegHooked ) UnhookRegistry();

	// delete the symbolic link for our device
	RtlInitUnicodeString( &deviceLinkUnicodeString, deviceLinkBuffer );
	IoDeleteSymbolicLink( &deviceLinkUnicodeString );

	// Delete the device object
	IoDeleteDevice( DriverObject->DeviceObject );

	DbgPrint(("Regmon.SYS: deleted devices\n"));

	// now we can free any memory we have outstanding
	RegmonHashCleanup();
	RegmonFreeStore();

	DbgPrint(("Regmon.SYS: freed memory\n"));
}


//----------------------------------------------------------------------
//
// DriverEntry
//
// Installable driver initialization. Here we just set ourselves up.
//
//----------------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath )
{
	NTSTATUS                ntStatus;
	WCHAR                   deviceNameBuffer[]  = L"\\Device\\Regmon";
	UNICODE_STRING          deviceNameUnicodeString;
	WCHAR                   deviceLinkBuffer[]  = L"\\DosDevices\\Regmon";
	UNICODE_STRING          deviceLinkUnicodeString;    
	int                     i;

	DbgPrint (("Regmon.SYS: entering DriverEntry\n"));

    	// get NT build number
	if( *NtBuildNumber > 1314 ) {
	  // NT 4.0 > beta 2
	  IsNT40 = TRUE;
	  nttable = nt40;
	} else if( *NtBuildNumber == 1314) {
	  // NT 4.0 beta 2
	  IsNT40 = TRUE;
	  nttable = nt40;			
	} else {
	  // 3.51 system
	  IsNT40 = FALSE;
	  nttable = nt351;			
	}

	// setup our name
	RtlInitUnicodeString (&deviceNameUnicodeString,
			  deviceNameBuffer );

	// set up the device used for GUI communications
	ntStatus = IoCreateDevice ( DriverObject,
			       0,
			       &deviceNameUnicodeString,
			       FILE_DEVICE_REGMON,
			       0,
			       TRUE,
			       &GUIDevice );
	if (NT_SUCCESS(ntStatus)) {
		// Create a symbolic link that the GUI can specify to gain access
		// to this driver/device
		RtlInitUnicodeString (&deviceLinkUnicodeString,
			      deviceLinkBuffer );
		ntStatus = IoCreateSymbolicLink (&deviceLinkUnicodeString,
					 &deviceNameUnicodeString );
		if (!NT_SUCCESS(ntStatus))
			DbgPrint (("Regmon.SYS: IoCreateSymbolicLink failed\n"));

		// Create dispatch points for all routines that must be handled
		DriverObject->MajorFunction[IRP_MJ_CREATE]          =
		DriverObject->MajorFunction[IRP_MJ_CLOSE]           =
		DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = RegmonDispatch;
		DriverObject->DriverUnload                          = RegmonUnload;
	}
	if (!NT_SUCCESS(ntStatus)) {
		DbgPrint(("Regmon: Failed to create our device!\n"));
		// Something went wrong, so clean up (free resources etc)
		if( GUIDevice ) IoDeleteDevice( GUIDevice );
		return ntStatus;
	}

	// initialize our semaphores
	MUTEX_INIT( StoreMutex );
	MUTEX_INIT( HashMutex );
	MUTEX_INIT( RegMutex );

	// initialize rootkey lengths
	for( i = 0; i < NUMROOTKEYS; i++ )
		RootKey[i].RootNameLen = strlen( RootKey[i].RootName );

	// pointer to system table is global export in NTOSRKNL.EXE
	ServiceTable = (PSRVTABLE) ((PULONG) KeServiceDescriptorTable );
	DbgPrint(("Hookregistry: Servicetable: %x\n", ServiceTable ));

	// We have to always have two buffers available to avoid deadlock due to reentrancy
	Store   = ExAllocatePool( NonPagedPool, sizeof(*Store) );
	if ( !Store )
		return STATUS_INSUFFICIENT_RESOURCES;
	Store->Len  = 0;
	Store->Next = NULL;
	NumStore = 1;

	return ntStatus;
}
    
